@@ -3,13 +3,18 @@ from concurrent import futures
 import grpc
 import base64
 from collections import defaultdict
+import os
+import queue
+import pickle
 
+import threading
 from typing import Any
 from typing import Dict
 from typing import Set
 from typing import Optional
-
+from typing import Callable
 from ray import cloudpickle
+from ray.job_config import JobConfig
 import ray
 import ray.state
 import ray.core.generated.ray_client_pb2 as ray_client_pb2
@@ -17,26 +22,83 @@ import ray.core.generated.ray_client_pb2_grpc as ray_client_pb2_grpc
 import time
 import inspect
 import json
+from ray.util.client.common import (ClientServerHandle, GRPC_OPTIONS,
+                                    CLIENT_SERVER_MAX_THREADS)
+from ray.util.client.server.proxier import serve_proxier
 from ray.util.client.server.server_pickler import convert_from_arg
 from ray.util.client.server.server_pickler import dumps_from_server
 from ray.util.client.server.server_pickler import loads_from_client
 from ray.util.client.server.dataservicer import DataServicer
 from ray.util.client.server.logservicer import LogstreamServicer
-from ray.util.client.server.server_stubs import current_remote
+from ray.util.client.server.server_stubs import current_server
+from ray.ray_constants import env_integer
+from ray.util.placement_group import PlacementGroup
 from ray._private.client_mode_hook import disable_client_hook
 
 logger = logging.getLogger(__name__)
 
+TIMEOUT_FOR_SPECIFIC_SERVER_S = env_integer("TIMEOUT_FOR_SPECIFIC_SERVER_S",
+                                            30)
+
 
 class RayletServicer(ray_client_pb2_grpc.RayletDriverServicer):
-    def __init__(self):
+    def __init__(self, ray_connect_handler: Callable):
+        """Construct a raylet service
+
+        Args:
+           ray_connect_handler (Callable): Function to connect to ray cluster
+        """
+        # Stores client_id -> (ref_id -> ObjectRef)
         self.object_refs: Dict[str, Dict[bytes, ray.ObjectRef]] = defaultdict(
             dict)
+        # Stores client_id -> (client_ref_id -> ref_id (in self.object_refs))
+        self.client_side_ref_map: Dict[str, Dict[bytes, bytes]] = defaultdict(
+            dict)
         self.function_refs = {}
         self.actor_refs: Dict[bytes, ray.ActorHandle] = {}
         self.actor_owners: Dict[str, Set[bytes]] = defaultdict(set)
         self.registered_actor_classes = {}
-        self._current_function_stub = None
+        self.named_actors = set()
+        self.state_lock = threading.Lock()
+        self.ray_connect_handler = ray_connect_handler
+
+    def Init(self, request, context=None) -> ray_client_pb2.InitResponse:
+        if request.job_config:
+            job_config = pickle.loads(request.job_config)
+            job_config.client_job = True
+        else:
+            job_config = None
+        current_job_config = None
+        with disable_client_hook():
+            if ray.is_initialized():
+                worker = ray.worker.global_worker
+                current_job_config = worker.core_worker.get_job_config()
+            else:
+                self.ray_connect_handler(job_config)
+        if job_config is None:
+            return ray_client_pb2.InitResponse()
+        job_config = job_config.get_proto_job_config()
+        # If the server has been initialized, we need to compare whether the
+        # runtime env is compatible.
+        if current_job_config and \
+           set(job_config.runtime_env.uris) != set(
+                current_job_config.runtime_env.uris) and \
+                len(job_config.runtime_env.uris) > 0:
+            return ray_client_pb2.InitResponse(
+                ok=False,
+                msg="Runtime environment doesn't match "
+                f"request one {job_config.runtime_env.uris} "
+                f"current one {current_job_config.runtime_env.uris}")
+        return ray_client_pb2.InitResponse(ok=True)
+
+    def PrepRuntimeEnv(self, request,
+                       context=None) -> ray_client_pb2.PrepRuntimeEnvResponse:
+        job_config = ray.worker.global_worker.core_worker.get_job_config()
+        try:
+            self._prepare_runtime_env(job_config.runtime_env)
+        except Exception as e:
+            raise grpc.RpcError(f"Prepare runtime env failed with {e}")
+        return ray_client_pb2.PrepRuntimeEnvResponse()
 
     def KVPut(self, request, context=None) -> ray_client_pb2.KVPutResponse:
         with disable_client_hook():
@@ -60,6 +122,13 @@ class RayletServicer(ray_client_pb2_grpc.RayletDriverServicer):
                 request.prefix)
         return ray_client_pb2.KVListResponse(keys=keys)
 
+    def KVExists(self, request,
+                 context=None) -> ray_client_pb2.KVExistsResponse:
+        with disable_client_hook():
+            exists = ray.experimental.internal_kv._internal_kv_exists(
+                request.key)
+        return ray_client_pb2.KVExistsResponse(exists=exists)
+
     def ClusterInfo(self, request,
                     context=None) -> ray_client_pb2.ClusterInfoResponse:
         resp = ray_client_pb2.ClusterInfoResponse()
@@ -89,8 +158,10 @@ class RayletServicer(ray_client_pb2_grpc.RayletDriverServicer):
                 rtc = ray.get_runtime_context()
                 ctx.job_id = rtc.job_id.binary()
                 ctx.node_id = rtc.node_id.binary()
+                ctx.namespace = rtc.namespace
                 ctx.capture_client_tasks = \
                     rtc.should_capture_child_tasks_in_placement_group
+                ctx.runtime_env = json.dumps(rtc.runtime_env)
             resp.runtime_context.CopyFrom(ctx)
         else:
             with disable_client_hook():
@@ -98,34 +169,49 @@ class RayletServicer(ray_client_pb2_grpc.RayletDriverServicer):
         return resp
 
     def _return_debug_cluster_info(self, request, context=None) -> str:
+        """Handle ClusterInfo requests that only return a json blob."""
         data = None
         if request.type == ray_client_pb2.ClusterInfoType.NODES:
             data = ray.nodes()
         elif request.type == ray_client_pb2.ClusterInfoType.IS_INITIALIZED:
             data = ray.is_initialized()
+        elif request.type == ray_client_pb2.ClusterInfoType.TIMELINE:
+            data = ray.timeline()
+        elif request.type == ray_client_pb2.ClusterInfoType.PING:
+            data = {}
         else:
             raise TypeError("Unsupported cluster info type")
         return json.dumps(data)
 
     def release(self, client_id: str, id: bytes) -> bool:
-        if client_id in self.object_refs:
-            if id in self.object_refs[client_id]:
-                logger.debug(f"Releasing object {id.hex()} for {client_id}")
-                del self.object_refs[client_id][id]
-                return True
+        with self.state_lock:
+            if client_id in self.object_refs:
+                if id in self.object_refs[client_id]:
+                    logger.debug(
+                        f"Releasing object {id.hex()} for {client_id}")
+                    del self.object_refs[client_id][id]
+                    return True
 
-        if client_id in self.actor_owners:
-            if id in self.actor_owners[client_id]:
-                logger.debug(f"Releasing actor {id.hex()} for {client_id}")
-                del self.actor_refs[id]
-                self.actor_owners[client_id].remove(id)
-                return True
+            if client_id in self.actor_owners:
+                if id in self.actor_owners[client_id]:
+                    logger.debug(f"Releasing actor {id.hex()} for {client_id}")
+                    self.actor_owners[client_id].remove(id)
+                    if self._can_remove_actor_ref(id):
+                        logger.debug(f"Deleting reference to actor {id.hex()}")
+                        del self.actor_refs[id]
+                    return True
 
-        return False
+            return False
 
     def release_all(self, client_id):
-        self._release_objects(client_id)
-        self._release_actors(client_id)
+        with self.state_lock:
+            self._release_objects(client_id)
+            self._release_actors(client_id)
+
+    def _can_remove_actor_ref(self, actor_id_bytes):
+        no_owner = not any(actor_id_bytes in actor_list
+                           for actor_list in self.actor_owners.values())
+        return no_owner and actor_id_bytes not in self.named_actors
 
     def _release_objects(self, client_id):
         if client_id not in self.object_refs:
@@ -133,16 +219,23 @@ class RayletServicer(ray_client_pb2_grpc.RayletDriverServicer):
             return
         count = len(self.object_refs[client_id])
         del self.object_refs[client_id]
+        if client_id in self.client_side_ref_map:
+            del self.client_side_ref_map[client_id]
         logger.debug(f"Released all {count} objects for client {client_id}")
 
     def _release_actors(self, client_id):
         if client_id not in self.actor_owners:
             logger.debug(f"Releasing client with no actors: {client_id}")
+            return
+
         count = 0
-        for id_bytes in self.actor_owners[client_id]:
+        actors_to_remove = self.actor_owners.pop(client_id)
+        for id_bytes in actors_to_remove:
             count += 1
-            del self.actor_refs[id_bytes]
-        del self.actor_owners[client_id]
+            if self._can_remove_actor_ref(id_bytes):
+                logger.debug(f"Deleting reference to actor {id_bytes.hex()}")
+                del self.actor_refs[id_bytes]
+
         logger.debug(f"Released all {count} actors for client: {client_id}")
 
     def Terminate(self, req, context=None):
@@ -170,6 +263,48 @@ class RayletServicer(ray_client_pb2_grpc.RayletDriverServicer):
                 "terminate_type")
         return ray_client_pb2.TerminateResponse(ok=True)
 
+    def _async_get_object(
+            self,
+            request,
+            client_id: str,
+            req_id: int,
+            result_queue: queue.Queue,
+            context=None) -> Optional[ray_client_pb2.GetResponse]:
+        """Attempts to schedule a callback to push the GetResponse to the
+        main loop when the desired object is ready. If there is some failure
+        in scheduling, a GetResponse will be immediately returned.
+        """
+        if request.id not in self.object_refs[client_id]:
+            return ray_client_pb2.GetResponse(valid=False)
+        try:
+            object_ref = self.object_refs[client_id][request.id]
+            logger.debug("async get: %s" % object_ref)
+            with disable_client_hook():
+
+                def send_get_response(result: Any) -> None:
+                    """Pushes a GetResponse to the main DataPath loop to send
+                    to the client. This is called when the object is ready
+                    on the server side."""
+                    try:
+                        serialized = dumps_from_server(result, client_id, self)
+                        get_resp = ray_client_pb2.GetResponse(
+                            valid=True, data=serialized)
+                    except Exception as e:
+                        get_resp = ray_client_pb2.GetResponse(
+                            valid=False, error=cloudpickle.dumps(e))
+
+                    resp = ray_client_pb2.DataResponse(
+                        get=get_resp, req_id=req_id)
+                    resp.req_id = req_id
+
+                    result_queue.put(resp)
+
+                object_ref._on_completed(send_get_response)
+                return None
+        except Exception as e:
+            return ray_client_pb2.GetResponse(
+                valid=False, error=cloudpickle.dumps(e))
+
     def GetObject(self, request, context=None):
         return self._get_object(request, "", context)
 
@@ -205,12 +340,21 @@ class RayletServicer(ray_client_pb2_grpc.RayletDriverServicer):
               delete this reference.
             context: gRPC context.
         """
-        obj = loads_from_client(request.data, self)
-        with disable_client_hook():
-            objectref = ray.put(obj)
+        try:
+            obj = loads_from_client(request.data, self)
+            with disable_client_hook():
+                objectref = ray.put(obj)
+        except Exception as e:
+            logger.exception("Put failed:")
+            return ray_client_pb2.PutResponse(
+                id=b"", valid=False, error=cloudpickle.dumps(e))
+
         self.object_refs[client_id][objectref.binary()] = objectref
+        if len(request.client_ref_id) > 0:
+            self.client_side_ref_map[client_id][
+                request.client_ref_id] = objectref.binary()
         logger.debug("put: %s" % objectref)
-        return ray_client_pb2.PutResponse(id=objectref.binary())
+        return ray_client_pb2.PutResponse(id=objectref.binary(), valid=True)
 
     def WaitObject(self, request, context=None) -> ray_client_pb2.WaitResponse:
         object_refs = []
@@ -299,7 +443,7 @@ class RayletServicer(ray_client_pb2_grpc.RayletDriverServicer):
         opts = decode_options(task.options)
         if opts is not None:
             remote_class = remote_class.options(**opts)
-        with current_remote(remote_class):
+        with current_server(self):
             actor = remote_class.remote(*arglist, **kwargs)
         self.actor_refs[actor._actor_id.binary()] = actor
         self.actor_owners[task.client_id].add(actor._actor_id.binary())
@@ -315,7 +459,7 @@ class RayletServicer(ray_client_pb2_grpc.RayletDriverServicer):
         opts = decode_options(task.options)
         if opts is not None:
             remote_func = remote_func.options(**opts)
-        with current_remote(remote_func):
+        with current_server(self):
             output = remote_func.remote(*arglist, **kwargs)
         ids = self.unify_and_track_outputs(output, task.client_id)
         return ray_client_pb2.ClientTaskTicket(return_ids=ids)
@@ -325,8 +469,10 @@ class RayletServicer(ray_client_pb2_grpc.RayletDriverServicer):
                               context=None) -> ray_client_pb2.ClientTaskTicket:
         assert len(task.payload_id) == 0
         actor = ray.get_actor(task.name)
-        self.actor_refs[actor._actor_id.binary()] = actor
-        self.actor_owners[task.client_id].add(actor._actor_id.binary())
+        bin_actor_id = actor._actor_id.binary()
+        self.actor_refs[bin_actor_id] = actor
+        self.actor_owners[task.client_id].add(bin_actor_id)
+        self.named_actors.add(bin_actor_id)
         return ray_client_pb2.ClientTaskTicket(
             return_ids=[actor._actor_id.binary()])
 
@@ -340,6 +486,15 @@ class RayletServicer(ray_client_pb2_grpc.RayletDriverServicer):
             kwargout[k] = convert_from_arg(kwarg_map[k], self)
         return argout, kwargout
 
+    def _prepare_runtime_env(self, job_runtime_env):
+        """Download runtime environment to local node"""
+        uris = job_runtime_env.uris
+        from ray._private import runtime_env
+        with disable_client_hook():
+            working_dir = runtime_env.ensure_runtime_env_setup(uris)
+            if working_dir:
+                os.chdir(working_dir)
+
     def lookup_or_register_func(
             self, id: bytes, client_id: str,
             options: Optional[Dict]) -> ray.remote_function.RemoteFunction:
@@ -404,25 +559,29 @@ def decode_options(
         return None
     opts = json.loads(options.json_options)
     assert isinstance(opts, dict)
+
+    if opts.get("placement_group", None):
+        # Placement groups in Ray client options are serialized as dicts.
+        # Convert the dict to a PlacementGroup.
+        opts["placement_group"] = PlacementGroup.from_dict(
+            opts["placement_group"])
+
     return opts
 
 
-_current_servicer: Optional[RayletServicer] = None
+def serve(connection_str, ray_connect_handler=None):
+    def default_connect_handler(job_config: JobConfig = None):
+        with disable_client_hook():
+            if not ray.is_initialized():
+                return ray.init(job_config=job_config)
 
-
-# Used by tests to peek inside the servicer
-def _get_current_servicer():
-    global _current_servicer
-    return _current_servicer
-
-
-def serve(connection_str):
-    server = grpc.server(futures.ThreadPoolExecutor(max_workers=10))
-    task_servicer = RayletServicer()
+    ray_connect_handler = ray_connect_handler or default_connect_handler
+    server = grpc.server(
+        futures.ThreadPoolExecutor(max_workers=CLIENT_SERVER_MAX_THREADS),
+        options=GRPC_OPTIONS)
+    task_servicer = RayletServicer(ray_connect_handler)
     data_servicer = DataServicer(task_servicer)
     logs_servicer = LogstreamServicer()
-    global _current_servicer
-    _current_servicer = task_servicer
     ray_client_pb2_grpc.add_RayletDriverServicer_to_server(
         task_servicer, server)
     ray_client_pb2_grpc.add_RayletDataStreamerServicer_to_server(
@@ -430,16 +589,32 @@ def serve(connection_str):
     ray_client_pb2_grpc.add_RayletLogStreamerServicer_to_server(
         logs_servicer, server)
     server.add_insecure_port(connection_str)
+    current_handle = ClientServerHandle(
+        task_servicer=task_servicer,
+        data_servicer=data_servicer,
+        logs_servicer=logs_servicer,
+        grpc_server=server,
+    )
     server.start()
-    return server
+    return current_handle
 
 
 def init_and_serve(connection_str, *args, **kwargs):
     with disable_client_hook():
         # Disable client mode inside the worker's environment
         info = ray.init(*args, **kwargs)
-    server = serve(connection_str)
-    return (server, info)
+
+    def ray_connect_handler(job_config=None):
+        # Ray client will disconnect from ray when
+        # num_clients == 0.
+        if ray.is_initialized():
+            return info
+        else:
+            return ray.init(job_config=job_config, *args, **kwargs)
+
+    server_handle = serve(
+        connection_str, ray_connect_handler=ray_connect_handler)
+    return (server_handle, info)
 
 
 def shutdown_with_server(server, _exiting_interpreter=False):
@@ -448,13 +623,49 @@ def shutdown_with_server(server, _exiting_interpreter=False):
         ray.shutdown(_exiting_interpreter)
 
 
+def create_ray_handler(redis_address, redis_password):
+    def ray_connect_handler(job_config: JobConfig = None):
+        if redis_address:
+            if redis_password:
+                ray.init(
+                    address=redis_address,
+                    _redis_password=redis_password,
+                    job_config=job_config)
+            else:
+                ray.init(address=redis_address, job_config=job_config)
+        else:
+            ray.init(job_config=job_config)
+
+    return ray_connect_handler
+
+
+def try_create_redis_client(args):
+    if "redis-address" not in args:
+        possible = ray._private.services.find_redis_address()
+        if len(possible) != 1:
+            return None
+        address = possible.pop()
+    else:
+        address = args["redis-address"]
+    if args.redis_password is None:
+        password = ray.ray_constants.REDIS_DEFAULT_PASSWORD
+    else:
+        password = args.redis_password
+    return ray._private.services.create_redis_client(address, password)
+
+
 def main():
     import argparse
     parser = argparse.ArgumentParser()
     parser.add_argument(
         "--host", type=str, default="0.0.0.0", help="Host IP to bind to")
     parser.add_argument(
-        "-p", "--port", type=int, default=50051, help="Port to bind to")
+        "-p", "--port", type=int, default=10001, help="Port to bind to")
+    parser.add_argument(
+        "--mode",
+        type=str,
+        choices=["proxy", "legacy", "specific-server"],
+        default="proxy")
     parser.add_argument(
         "--redis-address",
         required=False,
@@ -467,21 +678,53 @@ def main():
         help="Password for connecting to Redis")
     args = parser.parse_args()
     logging.basicConfig(level="INFO")
-    if args.redis_address:
-        if args.redis_password:
-            ray.init(
-                address=args.redis_address,
-                _redis_password=args.redis_password)
-        else:
-            ray.init(address=args.redis_address)
-    else:
-        ray.init()
+
+    # This redis client is used for health checking. We can't use `internal_kv`
+    # because it requires `ray.init` to be called, which only connect handlers
+    # should do.
+    redis_client = None
+
+    ray_connect_handler = create_ray_handler(args.redis_address,
+                                             args.redis_password)
+
     hostport = "%s:%d" % (args.host, args.port)
     logger.info(f"Starting Ray Client server on {hostport}")
-    server = serve(hostport)
+    if args.mode == "proxy":
+        server = serve_proxier(hostport, args.redis_address)
+    else:
+        server = serve(hostport, ray_connect_handler)
+
     try:
+        idle_checks_remaining = TIMEOUT_FOR_SPECIFIC_SERVER_S
         while True:
-            time.sleep(1000)
+            health_report = {
+                "time": time.time(),
+            }
+
+            try:
+                if not redis_client:
+                    redis_client = try_create_redis_client(args)
+                redis_client.hset("healthcheck:ray_client_server", "value",
+                                  json.dumps(health_report))
+            except Exception as e:
+                logger.error(f"[{args.mode}] Failed to put health check "
+                             f"on {args.redis_address}")
+                logger.exception(e)
+
+            time.sleep(1)
+            if args.mode == "specific-server":
+                if server.data_servicer.num_clients > 0:
+                    idle_checks_remaining = TIMEOUT_FOR_SPECIFIC_SERVER_S
+                else:
+                    idle_checks_remaining -= 1
+                if idle_checks_remaining == 0:
+                    raise KeyboardInterrupt()
+                if (idle_checks_remaining % 5 == 0 and idle_checks_remaining !=
+                        TIMEOUT_FOR_SPECIFIC_SERVER_S):
+                    logger.info(
+                        f"{idle_checks_remaining} idle checks before shutdown."
+                    )
+
     except KeyboardInterrupt:
         server.stop(0)
 
